"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const hasbin = require("hasbin");
const leanclient = require("lean-client-js-node");
const lean_client_js_node_1 = require("lean-client-js-node");
const os_1 = require("os");
const path_1 = require("path");
const username = require("username");
const vscode_1 = require("vscode");
const util_1 = require("./util");
// A global channel for storing the contents of stderr.
let stderrOutput;
// A class for interacting with the Lean server protocol.
class Server extends leanclient.Server {
    constructor() {
        super(null); // TODO(gabriel): add support to lean-client-js
        this.statusChanged = new util_1.LowPassFilter(300);
        this.restarted = new lean_client_js_node_1.Event();
        this.messages = [];
        this.attachEventHandlers();
    }
    connect() {
        try {
            this.messages = [];
            const config = vscode_1.workspace.getConfiguration('lean');
            // TODO(gabriel): unset LEAN_PATH environment variable
            this.executablePath = config.get('executablePath') || 'lean';
            if (this.executablePath === 'lean') {
                if (!hasbin.sync('lean')) {
                    // Let's try a little harder!
                    let elanLean = path_1.resolve(os_1.homedir(), '.elan', 'bin', 'lean');
                    if (fs_1.existsSync(elanLean)) {
                        this.executablePath = elanLean;
                    }
                    else if (process.platform === 'win32') {
                        elanLean = path_1.resolve('C:', 'msys64', 'home', username.sync(), '.elan', 'bin', 'lean');
                        if (fs_1.existsSync(elanLean)) {
                            this.executablePath = elanLean;
                        }
                    }
                }
            }
            this.workingDirectory = vscode_1.workspace.rootPath;
            this.options = config.get('extraOptions') || [];
            this.hasLean = true; // we don't need to ask to install lean if we got here
            this.options.push('-M');
            this.options.push('' + config.get('memoryLimit'));
            this.options.push('-T');
            this.options.push('' + config.get('timeLimit'));
            this.transport = new lean_client_js_node_1.ProcessTransport(this.executablePath, this.workingDirectory, this.options);
            super.connect();
            this.restarted.fire(null);
        }
        catch (e) {
            this.requestRestart(`Lean: ${e}`);
        }
    }
    attachEventHandlers() {
        // When attaching event handlers ensure the global error log is clear.
        stderrOutput = stderrOutput || vscode_1.window.createOutputChannel('Lean: Server Errors');
        stderrOutput.clear();
        this.error.on((e) => {
            switch (e.error) {
                case 'stderr':
                    stderrOutput.append(e.chunk);
                    stderrOutput.show();
                    break;
                case 'connect':
                    // json parsing errors
                    if (e.message.startsWith('cannot parse: ')) {
                        stderrOutput.append(e.message + '\n');
                        stderrOutput.show();
                        break;
                    }
                    this.requestRestart(`Lean: ${e.message}\n` +
                        'The lean.executablePath may be incorrect, make sure it is a valid Lean executable');
                    break;
                case 'unrelated':
                    vscode_1.window.showWarningMessage('Lean: ' + e.message);
                    break;
            }
            if (!this.alive()) {
                this.statusChanged.input({
                    isRunning: false,
                    numberOfTasks: 0,
                    stopped: true,
                    tasks: [],
                }, true);
            }
        });
        this.allMessages.on((msgs) => this.messages = msgs.msgs);
        this.tasks.on((curTasks) => this.statusChanged.input({
            isRunning: curTasks.is_running,
            numberOfTasks: curTasks.tasks.length,
            stopped: false,
            tasks: curTasks.tasks,
        }, curTasks.tasks.length === 0));
    }
    restart() {
        super.restart();
        stderrOutput.appendLine('----- user triggered restart -----');
    }
    installElan() {
        return __awaiter(this, void 0, void 0, function* () {
            if (vscode_1.workspace.getConfiguration('lean').get('executablePath') !== 'lean') {
                yield vscode_1.window.showErrorMessage("It looks like you've modified the `lean.executablePath` user setting.\n" +
                    'Please change it back to `lean` before installing elan.');
            }
            else {
                this.hasLean = true;
                const gitBashPath = 'C:\\Program Files\\Git\\bin\\bash.exe';
                const msysBashPath = 'C:\\msys64\\usr\\bin\\bash.exe';
                const terminalName = 'Lean installation via elan';
                const terminalOptions = { name: terminalName };
                if (process.platform === 'win32') {
                    if (fs_1.existsSync(gitBashPath)) {
                        terminalOptions.shellPath = gitBashPath;
                        terminalOptions.shellArgs = [];
                    }
                    else if (fs_1.existsSync(msysBashPath)) {
                        terminalOptions.shellPath = msysBashPath;
                        terminalOptions.shellArgs = ['--login', '-i'];
                    }
                    else {
                        yield vscode_1.window.showErrorMessage("You'll need to install a terminal (e.g. Git for Windows, or MSYS2)\n" +
                            `before we can install elan.`);
                        return;
                    }
                }
                const terminal = vscode_1.window.createTerminal(terminalOptions);
                // We register a listener, to restart the Lean extension once elan has finished.
                vscode_1.window.onDidCloseTerminal((t) => {
                    if (t.name === terminalName) {
                        this.restart();
                    }
                });
                // Now show the terminal and run elan.
                terminal.show();
                terminal.sendText('curl https://raw.githubusercontent.com/Kha/elan/master/elan-init.sh -sSf | sh && ' +
                    'echo && read -n 1 -s -r -p "Press any key to start Lean" && exit\n');
            }
        });
    }
    requestRestart(message, justWarning) {
        return __awaiter(this, void 0, void 0, function* () {
            const restartItem = 'Restart server';
            const installElanItem = 'Install Lean using elan';
            const showMsg = justWarning ? vscode_1.window.showWarningMessage : vscode_1.window.showErrorMessage;
            const item = this.hasLean ? yield showMsg(message, restartItem)
                : yield showMsg(message, restartItem, installElanItem);
            if (item === restartItem) {
                this.restart();
            }
            else if (item === installElanItem) {
                this.installElan();
            }
        });
    }
}
exports.Server = Server;
//# sourceMappingURL=server.js.map